#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sqlite.h>
#include <time.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>


#define MAXLINELENGTH 4000
#define PORT 5567
#define MESSAGETYPELENGTH 1
#define PAYLOADLENGTHLENGTH 2
#define WSNDBPATH "/mnt/state/adam/db/"
#define PVDB "/mnt/state/adam/db/pv.db"
#define SEPARATOR ";"
#define INNERSEPARATOR ","
#define CODEPROPBINARY "/sbin/codeprop"
#define CODEPROPTIMEOUT 5

#pragma pack(1)     /* set alignment to 1 byte boundary */
 
void handlingSocketCommunication(int sock);
void testfunction(void);
int createSocket(int domain, int type, int protocol);
void createServer(struct hostent **server, char *hostname);
void initServAddrServer(struct sockaddr_in * serv_addr, int portno);
void initServAddrClient(struct hostent **server, struct sockaddr_in *serv_addr, int portno);
void bindSocket(int sockfd, struct sockaddr_in *serv_addr);
void connectSocket(int sockfd, struct sockaddr_in *serv_addr);
void sendMessage(int sockfd, char *messagetype, char *payload);
void acceptSocket(int sockfd);
void parseStream(char* inbuffer, char* outbuffer);
int callback(void *result, int argc, char **argv, char **azColName);
char * getwsndb(char * wsndb);
int runSqlQuery(char *database, char *sqlquery, char *sqlresult);
int checkSqlResult(int resultErrorValue, char *result, char *sqlresult, int returnValueExpected);
void getPropertyValue(unsigned short sn_id, unsigned short prop_id, char *result);
void insertPropertyValue(unsigned short sn_id, unsigned short prop_id, char *value, char *result);
void updatePropertyValue(unsigned short sn_id, unsigned short prop_id, char *value, char *result);
void insertProperty(unsigned short prop_id, unsigned char type, char *description, char *result);
void getAllMeshNodes(char *result);
void getAllSensorNodes(char *result);
void getAllOwnSensorNodes(char *mn_id, char *result);
void getSensorNodeProperties(char *sn_id, char *result);
void insertSensorData(unsigned short sn_id, unsigned short prop_id, char *value, char *result);
void getSensorData(unsigned short sn_id, unsigned short prop_id, unsigned long time_from, unsigned long time_to, char *result);
void getCurrSensorData(unsigned short sn_id, unsigned short prop_id, char *result);
void getAllSensorData(char *result);
void getAllProgVersions(char *result);
void insertSensornodeConnection(unsigned short sn_id, unsigned short prop_id, char *result);
void getPvLink(unsigned short pv_id, char *result);
void updateSensorNodeImage(unsigned short sn_id, unsigned short pv_id, char *result);
void codepropError(char *result);
void codepropOk(char *result);
void codepropRestartError(char *result);
void databasequery(int sock, char *payload, char *outbuffer);
void send_msgtype1(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype2(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype3(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype4(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype5(char *completeMessage, int *completeMessageLength);
void send_msgtype6(char *completeMessage, int *completeMessageLength);
void send_msgtype7(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype8(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype9(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype10(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype11(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype12(char *completeMessage, int *completeMessageLength);
void send_msgtype13(char *completeMessage, int *completeMessageLength);
void send_msgtype14(char *payload, char *completeMessage, int *completeMessageLength);
void send_msgtype20(char *payload, char *completeMessage, int *completeMessageLength);
void recv_msgtype1(char *payload, char *outbuffer);
void recv_msgtype2(char *payload, char *outbuffer);
void recv_msgtype3(char *payload, char *outbuffer);
void recv_msgtype4(char *payload, char *outbuffer);
void recv_msgtype5(char *outbuffer);
void recv_msgtype6(char *outbuffer);
void recv_msgtype7(char *payload, char *outbuffer);
void recv_msgtype8(char *payload, char *outbuffer);
void recv_msgtype9(char *payload, char *outbuffer);
void recv_msgtype10(char *payload, char *outbuffer);
void recv_msgtype11(char *payload, char *outbuffer);
void recv_msgtype12(char *outbuffer);
void recv_msgtype13(char *outbuffer);
void recv_msgtype14(char *payload, char *outbuffer);
void recv_msgtype20(char *payload, char *outbuffer);
void executeSqlQuery(char *query, char *result);
void stripLastSemicolon(char *original, char *stripped);

void error(char *msg)
{
    perror(msg);
    exit(EXIT_FAILURE);
}

int main(int argc, char *argv[]) {
	int sockfd;
	struct sockaddr_in serv_addr;
  struct hostent *server;

  
  //act as server
  if (argc == 1) {
    sockfd = createSocket(AF_INET, SOCK_STREAM, 0);
    initServAddrServer(&serv_addr, PORT);
    bindSocket(sockfd, &serv_addr);
    listen(sockfd,5);
    printf("started marwis\n");

    while (1) {
      signal(SIGCHLD,SIG_IGN);
      acceptSocket(sockfd);
    }
    return 0; //we never get here
    
  //act as client
  } else if(argc >= 3) {
    sockfd = createSocket(AF_INET, SOCK_STREAM, 0);
  
    createServer(&server, argv[1]);
    initServAddrClient(&server, &serv_addr, PORT);
  
    connectSocket(sockfd, &serv_addr);
  
    sendMessage(sockfd, argv[2], argv[3]);
  
    return 0;
    
  } else {
    fprintf(stderr,"usage %s as server or %s <host> <messagetype> <payload> as client\n", argv[0], argv[0]);
    exit(EXIT_FAILURE);
  }
}

////////////////////////
//Server & Client Methodes
////////////////////////

//create the socket with integrated errorcheck
int createSocket(int domain, int type, int protocol) {
  int sockfd;
  sockfd = socket(domain, type, protocol);
  if (sockfd < 0) 
    error("ERROR opening socket");
  return sockfd;
}


////////////////////////
//Server Methodes
////////////////////////

//initialize the serveraddress acting as server
void initServAddrServer(struct sockaddr_in *serv_addr, int portno) {
  memset((char *) serv_addr, 0, sizeof(*serv_addr));
  serv_addr->sin_family = AF_INET;
  serv_addr->sin_addr.s_addr = INADDR_ANY;
  serv_addr->sin_port = htons(portno);
}

//bind socket on serveraddress
void bindSocket(int sockfd, struct sockaddr_in *serv_addr) {

  //avoiding "Address already in use"-error on binding
  int opt = 1;
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
    error("ERROR on setsockopt");
  }
  if (bind(sockfd, (struct sockaddr *) serv_addr,
      sizeof(*serv_addr)) < 0) 
    error("ERROR on binding");
}

//accept connections to socket
void acceptSocket(int sockfd) {
  int newsockfd, clilen, pid;
  struct sockaddr_in cli_addr;

  clilen = sizeof(cli_addr);
  newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
  if (newsockfd < 0) 
    error("ERROR on accept");
  pid = fork();
  if (pid < 0)
    error("ERROR on fork");
  if (pid == 0)  {
    close(sockfd);
    handlingSocketCommunication(newsockfd);
    exit(0);
  } else close(newsockfd);
}

//read from socket, do some stuff, write result to socket
void handlingSocketCommunication(int sock) {  
  int n;
  char inbuffer[MAXLINELENGTH+1];
  char outbuffer[MAXLINELENGTH+1]; 
  char payload[MAXLINELENGTH+1]; 
  memset(inbuffer, 0, MAXLINELENGTH+1);
  memset(outbuffer, 0, MAXLINELENGTH+1);
  memset(payload, 0, MAXLINELENGTH+1);
  unsigned char messagetype = 0;

  //read from socket
  n = read(sock,inbuffer,MAXLINELENGTH);
  if (n < 0) error("ERROR reading from socket");

  memcpy(&messagetype, inbuffer, MESSAGETYPELENGTH);
  memcpy(payload, inbuffer+MESSAGETYPELENGTH, MAXLINELENGTH-MESSAGETYPELENGTH+1);

  switch(messagetype) {
    case 1 : recv_msgtype1(payload, outbuffer); break;
    case 2 : recv_msgtype2(payload, outbuffer); break;
    case 3 : recv_msgtype3(payload, outbuffer); break;
    case 4 : recv_msgtype4(payload, outbuffer); break;
    case 5 : recv_msgtype5(outbuffer); break;
    case 6 : recv_msgtype6(outbuffer); break;
    case 7 : recv_msgtype7(payload, outbuffer); break;
    case 8 : recv_msgtype8(payload, outbuffer); break;
    case 9 : recv_msgtype9(payload, outbuffer); break;
    case 10 : recv_msgtype10(payload, outbuffer); break;
    case 11 : recv_msgtype11(payload, outbuffer); break;
    case 12 : recv_msgtype12(outbuffer); break;
    case 13 : recv_msgtype13(outbuffer); break;
    case 14 : recv_msgtype14(payload, outbuffer); break;
    case 20 : recv_msgtype20(payload, outbuffer); break;
    default  : printf("Unknown messagetype...\n");
  } 
  
  //write to socket
  n = write(sock,outbuffer,strlen(outbuffer));
  if (n < 0) error("ERROR writing to socket");
}

char * getwsndb(char * wsndb) {
	FILE *p = popen("/bin/hostname","r");
	if(p==NULL)
        {
          printf("Couldn't get the hostname\n");
          exit(1);
        }
	char buf[100];
	fgets(buf,100,p);
	strcpy (wsndb,WSNDBPATH);
 	strcat (wsndb,"wsn_");
 	strncat (wsndb,buf, (strlen(buf)-1));
  	strcat (wsndb,".db");
}


////////////////////////
//Messagetype Methodes
////////////////////////

void recv_msgtype1(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short first;
    unsigned short second;
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  getPropertyValue(pl->first, pl->second, outbuffer);
}

void recv_msgtype2(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    unsigned short first;
    unsigned short second;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  insertPropertyValue(pl->first, pl->second, pl->third, outbuffer);
}
void recv_msgtype3(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    unsigned short first;
    unsigned short second;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  updatePropertyValue(pl->first, pl->second, pl->third, outbuffer);
}
void recv_msgtype4(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    unsigned short first;
    char second;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  pl->first = ntohs(pl->first);
  //pl->second = ntohs(pl->second);
  
  insertProperty(pl->first, pl->second, pl->third, outbuffer);
}
void recv_msgtype5(char *outbuffer) {  
  getAllMeshNodes(outbuffer);
}
void recv_msgtype6(char *outbuffer) {  
  getAllSensorNodes(outbuffer);
}
void recv_msgtype7(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  
  getAllOwnSensorNodes(pl->third, outbuffer);
}
void recv_msgtype8(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  
  getSensorNodeProperties(pl->third, outbuffer);
}
void recv_msgtype9(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    unsigned short first;
    unsigned short second;
    char third[MAXLINELENGTH+1];
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  insertSensorData(pl->first, pl->second, pl->third, outbuffer);
}
void recv_msgtype10(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short length;
    unsigned short first;
    unsigned short second;
    unsigned long third;
    unsigned long fourth;
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->length = ntohs(pl->length);
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  pl->third = ntohl(pl->third);
  pl->fourth = ntohl(pl->fourth);
  
  getSensorData(pl->first, pl->second, pl->third, pl->fourth, outbuffer);
}
void recv_msgtype11(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short first;
    unsigned short second;
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  getCurrSensorData(pl->first, pl->second, outbuffer);
}
void recv_msgtype12(char *outbuffer) {  
  getAllSensorData(outbuffer);
}
void recv_msgtype13(char *outbuffer) {  
  getAllProgVersions(outbuffer);
}
void recv_msgtype14(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short first;
    unsigned short second;
  };
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  insertSensornodeConnection(pl->first, pl->second, outbuffer);
}
void recv_msgtype20(char *payload, char *outbuffer) {  
  struct PAYLOAD {
    unsigned short first;
    unsigned short second;
    unsigned short sensors[100];
  };
  
  int numberSensors = 0;
  int counter = 0;
  unsigned short sn_id = 0;
  
  struct PAYLOAD *pl = (struct PAYLOAD *) payload;
  pl->first = ntohs(pl->first);
  pl->second = ntohs(pl->second);
  
  numberSensors = pl->second;
  
  for(counter=0; counter<numberSensors; counter++) {
    updateSensorNodeImage(pl->first, ntohs(pl->sensors[counter]), outbuffer);
  }
  
}

////////////////////////
//Client Methodes
////////////////////////

//create the server with integrated errorcheck
void createServer(struct hostent **server, char *hostname) {
  *server = gethostbyname(hostname);
  if (server == NULL) {
    error("ERROR, no such host\n");
  }
}

//initialize the serveraddress acting as client
void initServAddrClient(struct hostent **server, struct sockaddr_in *serv_addr, int portno) {
  memset((char *) serv_addr, 0, sizeof(*serv_addr));
  serv_addr->sin_family = AF_INET;
  memcpy((char *)&(*serv_addr).sin_addr.s_addr, (char *)(*server)->h_addr, (*server)->h_length);
  serv_addr->sin_port = htons(portno);
}

//connect socket with server
void connectSocket(int sockfd, struct sockaddr_in *serv_addr) {
  if (connect(sockfd, (struct sockaddr *) serv_addr, sizeof(*serv_addr)) < 0) {
    error("ERROR connecting");
  }
}

//communication with server, printing answer from server
void sendMessage(int sockfd, char *messagetype, char *payload) {
  int n = 0;
  int completeMessageLength = 0;
  char msgtype = atoi(messagetype);
  
  char inbuffer[MAXLINELENGTH+1];
  char completeMessage[MAXLINELENGTH+1];
  memset(inbuffer, '\0', MAXLINELENGTH+1);
  memset(completeMessage, '\0', MAXLINELENGTH+1);
  memcpy(completeMessage, &msgtype, MESSAGETYPELENGTH);

  if(msgtype) {
    switch(msgtype) {
      case 1 : send_msgtype1(payload, completeMessage, &completeMessageLength); break;
      case 2 : send_msgtype2(payload, completeMessage, &completeMessageLength); break;
      case 3 : send_msgtype3(payload, completeMessage, &completeMessageLength); break;
      case 4 : send_msgtype4(payload, completeMessage, &completeMessageLength); break;
      case 5 : send_msgtype5(completeMessage, &completeMessageLength); break;
      case 6 : send_msgtype6(completeMessage, &completeMessageLength); break;
      case 7 : send_msgtype7(payload, completeMessage, &completeMessageLength); break;
      case 8 : send_msgtype8(payload, completeMessage, &completeMessageLength); break;
      case 9 : send_msgtype9(payload, completeMessage, &completeMessageLength); break;
      case 10 : send_msgtype10(payload, completeMessage, &completeMessageLength); break;
      case 11 : send_msgtype11(payload, completeMessage, &completeMessageLength); break;
      case 12 : send_msgtype12(completeMessage, &completeMessageLength); break;
      case 13 : send_msgtype13(completeMessage, &completeMessageLength); break;
      case 14 : send_msgtype14(payload, completeMessage, &completeMessageLength); break;
      case 20 : send_msgtype20(payload, completeMessage, &completeMessageLength); break;
      default  : printf("Unknown messagetype...\n"); return;
    }
    n = write(sockfd, completeMessage, completeMessageLength);

    if (n < 0) error("ERROR writing to socket");
    n = read(sockfd,inbuffer,MAXLINELENGTH+1);
    if (n < 0) error("ERROR reading from socket");
    printf("%s\n",inbuffer);
  }
}

void send_msgtype1(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+2, &second_4nw, 2);
  
  *completeMessageLength = MESSAGETYPELENGTH+2+2;
}
void send_msgtype2(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short thirdlength;
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  char *third = strtok(NULL, SEPARATOR);
  thirdlength = strlen(third);
  payloadlength = htons(2+2+thirdlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2, &second_4nw, 2);  
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2, third, thirdlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2+thirdlength;
}
void send_msgtype3(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short thirdlength;
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  char *third = strtok(NULL, SEPARATOR);
  thirdlength = strlen(third);
  payloadlength = htons(2+2+thirdlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2, &second_4nw, 2);  
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2, third, thirdlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2+thirdlength;
}
void send_msgtype4(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short thirdlength;
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  char second_4nw = atoi(strtok(NULL, SEPARATOR));
  
  char *third = strtok(NULL, SEPARATOR);
  thirdlength = strlen(third);
  payloadlength = htons(2+1+thirdlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2, &second_4nw, 1);  
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+1, third, thirdlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+1+thirdlength;
}
void send_msgtype5(char *completeMessage, int *completeMessageLength) {
  *completeMessageLength = MESSAGETYPELENGTH;
}
void send_msgtype6(char *completeMessage, int *completeMessageLength) {
  *completeMessageLength = MESSAGETYPELENGTH;
}
void send_msgtype7(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short firstlength;
  
  char *first = payload;
  firstlength = strlen(first);
  payloadlength = htons(firstlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, first, firstlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+firstlength;
}
void send_msgtype8(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short firstlength;
  
  char *first = payload;
  firstlength = strlen(first);
  payloadlength = htons(firstlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, first, firstlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+firstlength;
}
void send_msgtype9(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short thirdlength;
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  char *third = strtok(NULL, SEPARATOR);
  thirdlength = strlen(third);
  payloadlength = htons(2+2+thirdlength);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2, &second_4nw, 2);  
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2, third, thirdlength);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2+thirdlength;
}
void send_msgtype10(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short payloadlength;
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  unsigned long third_4nw = htonl(atol(strtok(NULL, SEPARATOR)));
  unsigned long fourth_4nw = htonl(atol(strtok(NULL, SEPARATOR)));
  payloadlength = htons(2+2+4+4);
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &payloadlength, PAYLOADLENGTHLENGTH);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2, &second_4nw, 2);  
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2, &third_4nw, 4);
  memcpy(completeMessage+MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2+4, &fourth_4nw, 4);
  
  *completeMessageLength = MESSAGETYPELENGTH+PAYLOADLENGTHLENGTH+2+2+4+4;
}
void send_msgtype11(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+2, &second_4nw, 2);
  
  *completeMessageLength = MESSAGETYPELENGTH+2+2;
}
void send_msgtype12(char *completeMessage, int *completeMessageLength) {
  *completeMessageLength = MESSAGETYPELENGTH;
}
void send_msgtype13(char *completeMessage, int *completeMessageLength) {
  *completeMessageLength = MESSAGETYPELENGTH;
}
void send_msgtype14(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short first_4nw = htons(atoi(strtok(payload, SEPARATOR)));
  unsigned short second_4nw = htons(atoi(strtok(NULL, SEPARATOR)));
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+2, &second_4nw, 2);
  
  *completeMessageLength = MESSAGETYPELENGTH+2+2;
}
void send_msgtype20(char *payload, char *completeMessage, int *completeMessageLength) {
  unsigned short first = atoi(strtok(payload, SEPARATOR));
  unsigned short first_4nw = htons(first);
  unsigned short second = atoi(strtok(NULL, SEPARATOR));
  unsigned short second_4nw = htons(second);
    
  unsigned short sensors[second];
  char *sensor = NULL;
  int counter = 0;
  
  sensor = strtok(NULL, SEPARATOR);
  while(sensor != NULL) {
    sensors[counter] = htons(atoi(sensor));
    sensor = strtok(NULL, SEPARATOR);
    counter++;
  }
  
  memcpy(completeMessage+MESSAGETYPELENGTH, &first_4nw, 2);
  memcpy(completeMessage+MESSAGETYPELENGTH+2, &second_4nw, 2);
  
  for(counter=0; counter<second; counter++) {
    memcpy(completeMessage+MESSAGETYPELENGTH+2+2+(counter*2), sensors+counter, 2);
  }
  
  *completeMessageLength = MESSAGETYPELENGTH+2+2+(counter*2);
  
}

//used by sqlite. this function processes each sqlite-result. output is res1, res2, ... , resn\n
int callback(void *result, int argc, char **argv, char **azColName) {
	int i;
	size_t len;
	//argc = number of columns
	for(i=0; i<argc; i++) {
		//argv[] = values in the columns
		len = MAXLINELENGTH - strlen(result);
		snprintf(result+strlen(result), len, "%s", argv[i] ? argv[i] : "NULL");
		if(i<argc-1) sprintf(result+strlen(result), INNERSEPARATOR);
	}
	len = MAXLINELENGTH - strlen(result);
	snprintf(result+strlen(result), len, SEPARATOR);
	return 0;
}



//open sqlite-database, execute query, close database
int runSqlQuery(char *database, char *sqlquery, char *sqlresult) {
	sqlite *db;
	char *zErrMsg = 0;
	int rc;
	char result[MAXLINELENGTH+1];
	memset(result, 0, MAXLINELENGTH+1);

	db = sqlite_open(database, 1, &zErrMsg);

	if(db==NULL) {
		fprintf(stderr, "Can't open database: %s\n", zErrMsg);
		sqlite_close(db);
		exit(EXIT_FAILURE);
	}

	rc = sqlite_exec(db, sqlquery, callback, result, &zErrMsg);
	if(rc!=SQLITE_OK) {
		fprintf(stderr, "SQL error: %s\n", zErrMsg);
//		sqlite_free(zErrMsg);
		return 1;
	}
	strncpy(sqlresult, result, MAXLINELENGTH);
	sqlite_close(db);
	return 0;
}

int checkSqlResult(int resultErrorValue, char *result, char *sqlresult, int returnValueExpected) {
	if(resultErrorValue == 1) {
		strncpy(result, "1", MAXLINELENGTH);
		return 1;
	} else {
		if(returnValueExpected == 1) {
			if(strcmp(sqlresult, "") == 0) {
				strncpy(result, "none", MAXLINELENGTH);
				return 1;
			} else {
				strncpy(result, sqlresult, MAXLINELENGTH);
				return 0;
			}
		} else {
			strncpy(result, "0", MAXLINELENGTH);
			return 0;
		}
	}
}

void executeSqlQuery(char *query, char *result) {
  int resultErrorValue;
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}


void getPropertyValue(unsigned short sn_id, unsigned short prop_id, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);

  snprintf(query, MAXLINELENGTH, "SELECT value FROM sensornodes WHERE sn_id='%d' AND prop_id='%d';",sn_id, prop_id);

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}




void insertPropertyValue(unsigned short sn_id, unsigned short prop_id, char *value, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);

  snprintf(query, MAXLINELENGTH, "INSERT INTO sensornodes (sn_id, prop_id, value, time) VALUES ('%d', '%d', '%s', '%d');", sn_id, prop_id, value, time(NULL));

//  printf("%s\n", query);

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 0);
}

void updatePropertyValue(unsigned short sn_id, unsigned short prop_id, char *value, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "UPDATE sensornodes SET value='%s', time='%d' WHERE sn_id='%d' AND prop_id='%d';", value, time(NULL), sn_id, prop_id);

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 0);
}

void insertProperty(unsigned short prop_id, unsigned char type, char *description, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char *stringtype;
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  switch(type) {
    case 0 : stringtype = "int"; break;
    case 1 : stringtype = "string"; break;
    default  : printf("Unknown propertytype...\n"); exit(1);
  } 
  
  snprintf(query, MAXLINELENGTH, "INSERT INTO properties (prop_id, description, type) VALUES ('%d', '%s', '%s');", prop_id, description, stringtype);

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 0);
}

void getAllMeshNodes(char *result) {
  int resultErrorValue;
  char query[] = "SELECT mn_id from Meshnodes ORDER BY mn_id ASC;";
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);	

  getwsndb(wsndb);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getAllSensorNodes(char *result) {
  int resultErrorValue;
  char query[] = "SELECT DISTINCT sn_id, value as mn_id from Sensornodes WHERE prop_id='24' ORDER BY sn_id ASC;";
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);	

  getwsndb(wsndb);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getAllOwnSensorNodes(char *mn_id, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "SELECT DISTINCT sn_id from Sensornodes WHERE prop_id='24' AND value='%s' ORDER BY sn_id ASC;", mn_id);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getSensorNodeProperties(char *sn_id, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "SELECT sensornodes.prop_id, properties.description, sensornodes.value from sensornodes LEFT JOIN properties USING (prop_id) WHERE sensornodes.sn_id='%s' ORDER BY sensornodes.prop_id ASC;", sn_id);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void insertSensorData(unsigned short sn_id, unsigned short prop_id, char *value, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "INSERT INTO sensors (sn_id, prop_id, value, time) VALUES ('%d', '%d', '%s', '%d');", sn_id, prop_id, value, time(NULL));

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 0);
}

void getSensorData(unsigned short sn_id, unsigned short prop_id, unsigned long time_from, unsigned long time_to, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "SELECT value, time from sensors WHERE sn_id='%u' AND prop_id='%u' AND time>='%lu' AND time<='%lu' ORDER BY time ASC;", sn_id, prop_id, time_from, time_to);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getCurrSensorData(unsigned short sn_id, unsigned short prop_id, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "SELECT value, time from sensors WHERE sn_id='%u' AND prop_id='%u' ORDER BY time DESC LIMIT 1;", sn_id, prop_id);
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getAllSensorData(char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);
	
  snprintf(query, MAXLINELENGTH, "SELECT sn_id, prop_id, value, time from sensors ORDER BY sn_id ASC, prop_id ASC, time ASC;");
	
  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void getAllProgVersions(char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);
	
  snprintf(query, MAXLINELENGTH, "SELECT * from programs ORDER BY pv_id ASC, name ASC;");
	
  resultErrorValue = runSqlQuery(PVDB, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
}

void insertSensornodeConnection(unsigned short sn_id, unsigned short prop_id, char *result) {
  int resultErrorValue;
  char query[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char wsndb[MAXLINELENGTH+1];
  memset(wsndb, 0, MAXLINELENGTH+1);
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);

  getwsndb(wsndb);

  snprintf(query, MAXLINELENGTH, "INSERT INTO sensornode_network (sn_id, sn_neighbor_id, time) VALUES ('%d', '%d', '%d');", sn_id, prop_id, time(NULL));

  resultErrorValue = runSqlQuery(wsndb, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 0);
}


void updateSensorNodeImage(unsigned short pv_id, unsigned short sn_id, char *result) {
  char pvlink_tmp[MAXLINELENGTH+1];
  char sn_ip_tmp[MAXLINELENGTH+1];
  char pvlink[MAXLINELENGTH+1];
  char sn_ip[MAXLINELENGTH+1];
  char sqlresult[MAXLINELENGTH+1];
  char codepropcommand [MAXLINELENGTH+1];
  int codepropresult = 0;
  int status;
  pid_t pid;
  
  getPvLink(pv_id, pvlink_tmp);  
  getPropertyValue(sn_id, 22, sn_ip_tmp);
  if(strcmp(sn_ip_tmp, "none") == 0) {
    //ip of sensornode not found in db
    return;
  }
  stripLastSemicolon(sn_ip_tmp, sn_ip);
  stripLastSemicolon(pvlink_tmp, pvlink);
  
  snprintf(codepropcommand, MAXLINELENGTH, "%s %s %s", CODEPROPBINARY, sn_ip, pvlink);
  
  printf("pvlink: %s\n", pvlink);
  printf("sn_ip: %s\n", sn_ip);
  printf("codepropcommand: %s\n", codepropcommand);
  
  pid = fork ();
  if (pid == 0)
  {
    execl ("/bin/sh", "/bin/sh", "-c", codepropcommand, NULL);
  }
  else if (pid < 0)
    status = -1;
  else
    /* This is the parent process.  Wait for the child to complete.  */
    sleep(CODEPROPTIMEOUT);
    kill(pid, SIGINT);
    if (waitpid (pid, &status, 0) != pid)
      status = -1;

  switch(status) {
    case -1 : codepropError(result); break;
    case 0 : codepropOk(result); break;
    case 1 : codepropRestartError(result); break;
    
    default  : printf("Unknown returnvalue from codeprop...\n");
  } 
}

void codepropError(char *result) {
  char *error = "Transmission failed";
  memcpy(result, error, strlen(error));
}

void codepropOk(char *result) {
  char *ok = "Update successful";
  memcpy(result, ok, strlen(ok));
}

void codepropRestartError(char *result) {
  char *restartError = "Transmission successful, Applicationstart failed";
  memcpy(result, restartError, strlen(restartError));
}

void getPvLink(unsigned short pv_id, char *result) {
  int resultErrorValue;
  char sqlresult[MAXLINELENGTH+1];
  char query[MAXLINELENGTH+1];
  memset(query, 0, MAXLINELENGTH+1);
  memset(sqlresult, 0, MAXLINELENGTH+1);
  
  snprintf(query, MAXLINELENGTH, "SELECT link from programs WHERE pv_id='%u';", pv_id);
  resultErrorValue = runSqlQuery(PVDB, query, sqlresult);
  checkSqlResult(resultErrorValue, result, sqlresult, 1);
  
}

void stripLastSemicolon(char *original, char *stripped) {
  memcpy(stripped, original, strlen(original)-1);
}
